use std::{
    collections::HashSet,
    ffi::{c_char, CStr, CString},
    ptr::null,
};

use serde::Deserialize;

// cracklib imports
extern "C" {
    /// Checks that the password is not based on a username or other credentials.
    fn FascistGecosUser(
        password: *const c_char,
        user: *const c_char,
        gecos: *const c_char,
    ) -> *mut c_char;
}

/// Password policy parameters specifying which checks will be performed.
#[derive(Debug, Default, Deserialize)]
pub struct PasswordPolicyParams {
    /// The minimum acceptable length for the new password
    pub min_len: Option<usize>,
    /// The minimum number of required classes of characters for the new password.
    pub min_char_classes: Option<usize>,
    /// Minimum number of different characters in a password.
    pub min_diff_chars: Option<usize>,
    /// The maximum length of a sequence of identical characters.
    pub max_repeat: Option<usize>,
    /// The maximum number of sequenced characters. Like 'abc'.
    pub max_sequence: Option<usize>,
    /// The maximum allowed sequence of characters of the same class.
    pub max_char_class_repeat: Option<usize>,
    /// Reject passwords based on username
    #[serde(default)]
    pub reject_username: bool,
}

/// Represent character classes.
#[derive(PartialEq, Eq, Hash, Debug)]
pub enum CharClass {
    Lowercase,
    Uppercase,
    Punctuation,
    Digit,
}

impl TryFrom<char> for CharClass {
    type Error = String;

    /// Try parse character and return the character class. The function is useful for
    /// calculating the general string class and class sequences inside string.
    ///
    /// # Examples
    ///
    /// ```
    /// use picodata::password_policy::CharClass;
    ///
    /// assert_eq!(CharClass::try_from('a'), Ok(CharClass::Lowercase));
    /// assert_eq!(CharClass::try_from('!'), Ok(CharClass::Punctuation));
    /// assert_eq!(CharClass::try_from('5'), Ok(CharClass::Digit));
    ///
    /// let password = "qWerty123!";
    /// let classes: Result<Vec<CharClass>, _> = password
    ///     .chars()
    ///     .map(|ch| CharClass::try_from(ch))
    ///     .collect();
    ///
    /// assert_eq!(classes, Ok(vec![
    ///     CharClass::Lowercase,
    ///     CharClass::Uppercase,
    ///     CharClass::Lowercase,
    ///     CharClass::Lowercase,
    ///     CharClass::Lowercase,
    ///     CharClass::Lowercase,
    ///     CharClass::Digit,
    ///     CharClass::Digit,
    ///     CharClass::Digit,
    ///     CharClass::Punctuation,
    /// ]));
    /// ```
    fn try_from(input: char) -> Result<Self, Self::Error> {
        match input {
            i if i.is_lowercase() => Ok(CharClass::Lowercase),
            i if i.is_uppercase() => Ok(CharClass::Uppercase),
            i if i.is_numeric() => Ok(CharClass::Digit),
            // Some symbols (like 码 or ₹) do not have lowercase or uppercase
            // versions. So we consider them just a lovercase.
            i if i.is_alphabetic() => Ok(CharClass::Lowercase),
            ' ' | '!' | '?' | '@' | '#' | '$' | '%' | '^' | '&' | '*' | '(' | ')' | '\'' | '"'
            | '`' | '_' | '-' | '+' | '~' | '>' | '<' | '=' | '.' | ',' | ':' | ';' => {
                Ok(CharClass::Punctuation)
            }
            _ => Err(format!("Symbol {input} not allowed for password")),
        }
    }
}

#[tarantool::proc]
pub fn proc_check_pass(
    password: &str,
    user: &str,
    params: PasswordPolicyParams,
) -> Result<(), String> {
    check(password, user, &params)
}

/// Check the password for a given set of optional parameters.
///
/// # Checks
///
/// - check if there are no whitespaces in the password
/// - check if there are no control characters
/// - check password length >= params.min_len
/// - number of char classes (lowercase, uppercase, digits, symbols) >= params.min_char_classes
/// - count unique characters >= params.min_diff_chars
/// - check the number of identical characters in the password <= params.max_repeat
/// - check sequences of identical characters <= params.max_sequence
/// - count sequence of same class symbols <= params.max_char_class_repeat
///
/// # Examples
///
/// ```
/// use picodata::password_policy::{PasswordPolicyParams, check};
///
/// let mut params = PasswordPolicyParams {
///     min_len: Some(10),
///     max_char_class_repeat: Some(100),
///     ..Default::default()
/// };
///
/// assert_eq!(check("mybrandnewpassword", "john", &params), Ok(()));
/// assert_eq!(
///     check("qwerty", "john", &params),
///     Err("password shorter than 10 characters".into())
/// );
///
/// params.min_len = Some(4);
/// params.reject_username = true;
///
/// assert_eq!(
///     check("nhoj", "john", &params),
///     Err("it is based on your username".into())
/// );
/// ```
pub fn check(password: &str, user: &str, params: &PasswordPolicyParams) -> Result<(), String> {
    let chars: Vec<char> = password.chars().collect();
    // create hashset with unique chars
    let unique_chars: HashSet<char> = chars.iter().copied().collect();

    // check password minimal length
    if let Some(min_len) = params.min_len {
        if password.len() < min_len {
            return Err(format!("password shorter than {min_len} characters"));
        }
    }

    // check that the password does not contain control characters
    if chars.iter().any(|ch| ch.is_control()) {
        return Err("password contains control characters".into());
    }

    // check the parameter is not greater than the maximum allowed class
    // the max class is the class that has all 4 character types
    // lowercase, uppercase, punctuation, digit
    if let Some(5..) = params.min_char_classes {
        return Err(
            "invalid value for 'min_char_classes' option: must not be greater than 4".into(),
        );
    }

    let password_class = unique_chars
        .iter()
        .map(|ch| CharClass::try_from(*ch))
        .collect::<Result<HashSet<CharClass>, _>>()?
        .len();

    // check that the password class is not lower than the minclass
    if let Some(min_char_classes) = params.min_char_classes {
        if password_class < min_char_classes {
            return Err(format!(
                "password does not contain enough character classes: \
                at least {min_char_classes} expected"
            ));
        }
    }

    // compare with min_diff_chars how many different characters we have
    // aabbccc -> not ok
    // abcdefa -> ok
    // aAbBcC -> ok
    if let Some(min_diff_chars) = params.min_diff_chars {
        if unique_chars.len() < min_diff_chars {
            return Err(format!(
                "password does not contain {min_diff_chars} different characters"
            ));
        }
    }

    // count repeated characters in password like aabbccc
    // if parameter value is 0, then the check is skipped
    if let Some(max_repeat @ 1..) = params.max_repeat {
        chars.windows(2).try_fold(1, |mut count, ch| {
            // count sequences of identical characters and following each other
            // in alphabetical order starting from chars[1]
            // arbcdt -> ok
            // aabbcc -> not ok
            // aAbBcC -> not ok
            if ch[0].to_lowercase().to_string() == ch[1].to_lowercase().to_string() {
                count += 1;
            } else {
                count = 1;
            }

            if count > max_repeat {
                return Err(format!(
                    "password must not contain more than {max_repeat} repeated characters"
                ));
            }

            Ok(count)
        })?;
    }

    // count sequenced characters is password like abcde
    // if parameter value is 0, then the check is skipped
    if let Some(max_sequence @ 1..) = params.max_sequence {
        // count for normal sequence like 'abcde'
        // reverse_count for reverse sequence like 'edcba'
        chars
            .windows(2)
            .try_fold((1, 1), |(mut count, mut reverse_count), window| {
                let mut left = [0; 2];

                window[0]
                    .to_lowercase()
                    .next()
                    .unwrap_or(window[0])
                    .encode_utf16(&mut left);

                let mut right = [0; 2];

                window[1]
                    .to_lowercase()
                    .next()
                    .unwrap_or(window[1])
                    .encode_utf16(&mut right);

                // left and right is utf-16 symbol or range number
                match (left[1], right[1]) {
                    // if left and right from utf-16 main range
                    // for example '㌀' and '㌁'
                    (0, 0) if left[0].checked_add(1).eq(&Some(right[0])) => {
                        count += 1;
                    }
                    // if left and right from utf-16 main range
                    // and are the reverse sequence
                    (0, 0) if left[0].checked_sub(1).eq(&Some(right[0])) => {
                        reverse_count += 1;
                    }
                    // if left and right from utf-16 extended range
                    // for example '🁷' and '🁸'
                    (_, _) if left[1].checked_add(1).eq(&Some(right[1])) => {
                        count += 1;
                    }
                    // if left and right from utf-16 extended range
                    // and are the reverse sequence
                    (_, _) if left[1].checked_sub(1).eq(&Some(right[1])) => {
                        reverse_count += 1;
                    }
                    _ => {
                        count = 1;
                        reverse_count = 1;
                    }
                }

                if usize::max(count, reverse_count) > max_sequence {
                    return Err(format!(
                        "password must not contain more than {max_sequence} sequenced characters"
                    ));
                }

                Ok((count, reverse_count))
            })?;
    }

    // count the number of repeated characters of the same class
    // if parameter value is 0, then the check is skipped
    if let Some(max_char_class_repeat @ 1..) = params.max_char_class_repeat {
        chars.windows(2).try_fold(1, |mut count, ch| {
            // if the class of the previous character and the current one match,
            // then we consider that this is a sequence of characters of the same
            match (CharClass::try_from(ch[0]), CharClass::try_from(ch[1])) {
                (Ok(left), Ok(right)) if left == right => {
                    count += 1;
                }
                _ => {
                    count = 1;
                }
            }

            if count > max_char_class_repeat {
                return Err(format!(
                    "the password must not contain sequences with more than \
                {max_char_class_repeat} characters of the same class"
                ));
            }

            Ok(count)
        })?;
    }

    // Checks that the password is not related to the username.
    if params.reject_username {
        let password_raw = CString::new(password.to_lowercase()).unwrap();
        let user_raw = CString::new(user.to_lowercase()).unwrap();

        let err_ptr = unsafe { FascistGecosUser(password_raw.as_ptr(), user_raw.as_ptr(), null()) };

        if !err_ptr.is_null() {
            let msg_str = unsafe { CStr::from_ptr(err_ptr) };
            return Err(msg_str.to_str().unwrap().into());
        }
    }

    Ok(())
}

#[cfg(test)]
mod test {
    use super::{check, PasswordPolicyParams};

    #[test]
    fn password_check() {
        ///////////////////////////////////////////////////////////////////
        // allow any passwords
        ///////////////////////////////////////////////////////////////////
        let allow_any = PasswordPolicyParams::default();

        assert_eq!(check("", "john", &allow_any), Ok(()));
        assert_eq!(check("a", "john", &allow_any), Ok(()));
        assert_eq!(check("1111", "john", &allow_any), Ok(()));

        ///////////////////////////////////////////////////////////////////
        // general tests
        ///////////////////////////////////////////////////////////////////
        let params = PasswordPolicyParams {
            min_len: Some(4),
            ..PasswordPolicyParams::default()
        };

        assert_eq!(check("xfzg", "john", &params), Ok(()));
        assert_eq!(
            check("xyz", "john", &params),
            Err("password shorter than 4 characters".into())
        );
        assert_eq!(
            check("xy\u{0004}z", "john", &params),
            Err("password contains control characters".into())
        );

        assert_eq!(check("密码", "Lao", &params), Ok(()));
        assert_eq!(check("1sfqtap", "john", &params), Ok(()));
        assert_eq!(check("كَلِمة سِرٍّ", "Nasir", &params), Ok(()));
        assert_eq!(
            check("م", "Nasir", &params),
            Err("password shorter than 4 characters".into())
        );

        ///////////////////////////////////////////////////////////////////
        // min_char_classes tests
        ///////////////////////////////////////////////////////////////////
        let mut params = PasswordPolicyParams {
            min_char_classes: Some(10),
            ..PasswordPolicyParams::default()
        };

        assert_eq!(
            check("xyzf", "john", &params),
            Err("invalid value for 'min_char_classes' option: must not be greater than 4".into())
        );

        params.min_char_classes = Some(2);

        assert_eq!(
            check("xyzf", "john", &params),
            Err("password does not contain enough character classes: at least 2 expected".into())
        );
        assert_eq!(check("xyzF", "john", &params), Ok(()));

        params.min_char_classes = Some(3);

        assert_eq!(
            check("xyzF", "john", &params),
            Err("password does not contain enough character classes: at least 3 expected".into())
        );
        assert_eq!(check("xyzF1", "john", &params), Ok(()));

        params.min_char_classes = Some(4);

        assert_eq!(
            check("xyzF1", "john", &params),
            Err("password does not contain enough character classes: at least 4 expected".into())
        );
        assert_eq!(check("xyzF1!", "john", &params), Ok(()));

        params.min_char_classes = Some(1);

        ///////////////////////////////////////////////////////////////////
        // min_diff_chars tests
        ///////////////////////////////////////////////////////////////////
        let mut params = PasswordPolicyParams {
            min_diff_chars: Some(3),
            ..PasswordPolicyParams::default()
        };

        assert_eq!(
            check("AbAAA", "john", &params),
            Err("password does not contain 3 different characters".into())
        );

        params.min_diff_chars = Some(5);

        assert_eq!(
            check("AbAcf", "john", &params),
            Err("password does not contain 5 different characters".into())
        );
        assert_eq!(check("abAcf", "john", &params), Ok(()));

        params.min_diff_chars = Some(3);

        ///////////////////////////////////////////////////////////////////
        // max_repeat tests
        ///////////////////////////////////////////////////////////////////
        let mut params = PasswordPolicyParams {
            max_repeat: Some(2),
            ..PasswordPolicyParams::default()
        };
        assert_eq!(check("arbcdt", "john", &params), Ok(()));
        assert_eq!(
            check("aabbbcc", "john", &params),
            Err("password must not contain more than 2 repeated characters".into())
        );
        assert_eq!(
            check("Abbbhf", "john", &params),
            Err("password must not contain more than 2 repeated characters".into())
        );
        assert_eq!(
            check("aAbBbCc", "john", &params),
            Err("password must not contain more than 2 repeated characters".into())
        );

        params.max_repeat = Some(1);

        assert_eq!(
            check("aAbBbCc", "john", &params),
            Err("password must not contain more than 1 repeated characters".into())
        );

        params.max_repeat = Some(0);

        // if parameter value is 0, then the check is skipped
        assert_eq!(check("AbBbhf", "john", &params), Ok(()));

        ///////////////////////////////////////////////////////////////////
        // max_sequence tests
        ///////////////////////////////////////////////////////////////////
        let mut params = PasswordPolicyParams {
            max_sequence: Some(0),
            ..PasswordPolicyParams::default()
        };

        // if parameter value is 0, then the check is skipped
        assert_eq!(check("abehg", "john", &params), Ok(()));

        params.max_sequence = Some(3);

        assert_eq!(
            check("abcdef", "john", &params),
            Err("password must not contain more than 3 sequenced characters".into())
        );
        assert_eq!(check("abehg", "john", &params), Ok(()));

        params.max_sequence = Some(0);

        // if parameter value is 0, then the check is skipped
        assert_eq!(check("abehg", "john", &params), Ok(()));

        params.max_sequence = Some(1);

        assert_eq!(check("xdsqt", "john", &params), Ok(()));

        params.max_sequence = Some(2);

        assert_eq!(check("abijcdxyop340189", "john", &params), Ok(()));
        assert_eq!(
            check("ฬอฮฯ", "Chan-Yeol", &params),
            Err("password must not contain more than 2 sequenced characters".into())
        );

        // reverse sequence
        assert_eq!(check("bajidcyxpo431098", "john", &params), Ok(()));

        params.max_sequence = Some(3);
        assert_eq!(check("abcba12321", "john", &params), Ok(()));
        assert_eq!(
            check("abcdcba", "john", &params),
            Err("password must not contain more than 3 sequenced characters".into())
        );
        assert_eq!(
            check("abjDcBa", "john", &params),
            Err("password must not contain more than 3 sequenced characters".into())
        );
        assert_eq!(
            check("abcSD@#dcba", "john", &params),
            Err("password must not contain more than 3 sequenced characters".into())
        );

        ///////////////////////////////////////////////////////////////////
        // max_char_class_repeat tests
        ///////////////////////////////////////////////////////////////////
        let mut params = PasswordPolicyParams {
            max_char_class_repeat: Some(0),
            ..PasswordPolicyParams::default()
        };

        assert_eq!(check("ahsr", "john", &params), Ok(()));

        params.max_char_class_repeat = Some(2);

        assert_eq!(
            check("ahsr", "john", &params),
            Err("the password must not contain sequences with more than \
                2 characters of the same class"
                .into())
        );
        assert_eq!(check("ahSr", "john", &params), Ok(()));

        params.max_char_class_repeat = Some(1);

        assert_eq!(
            check("ahSr", "john", &params),
            Err("the password must not contain sequences with more than \
                1 characters of the same class"
                .into())
        );
        assert_eq!(check("aHsR", "john", &params), Ok(()));

        ///////////////////////////////////////////////////////////////////
        // based on username
        ///////////////////////////////////////////////////////////////////
        let params = PasswordPolicyParams {
            reject_username: true,
            ..PasswordPolicyParams::default()
        };

        assert_eq!(
            check("john", "john", &params),
            Err("it is based on your username".into())
        );
        assert_eq!(
            check("nhoj", "john", &params),
            Err("it is based on your username".into())
        );
        assert_eq!(
            check("nHoJ", "john", &params),
            Err("it is based on your username".into())
        );
        assert_eq!(
            check("<john>", "john", &params),
            Err("it is based on your username".into())
        );
        assert_eq!(
            check("$up3rMAN", "superman", &params),
            Err("it is based on your username".into())
        );
        // this password is ok, because there are more non similar than are similar
        assert_eq!(check("->>j0hn===>", "john", &params), Ok(()));
        // non ascii symbols
        assert_eq!(
            check("先輩", "先輩", &params),
            Err("it is based on your username".into())
        );
    }
}
